@@ -1,7 +1,7 @@
 /*
  * media-factory.c - Media channel factory for SIP connection manager
  * Copyright (C) 2007-2008 Collabora Ltd.
- * Copyright (C) 2007-2008 Nokia Corporation
+ * Copyright (C) 2007-2010 Nokia Corporation
  *
  * This work is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -35,6 +35,11 @@
 #define DEBUG_FLAG TPSIP_DEBUG_CONNECTION
 #include "debug.h"
 
+typedef enum {
+  TPSIP_MEDIA_CHANNEL_CREATE_WITH_AUDIO = 1 << 0,
+  TPSIP_MEDIA_CHANNEL_CREATE_WITH_VIDEO = 1 << 1,
+} TpsipMediaChannelCreationFlags;
+
 static void channel_manager_iface_init (gpointer, gpointer);
 static void tpsip_media_factory_constructed (GObject *object);
 static void tpsip_media_factory_close_all (TpsipMediaFactory *fac);
@@ -255,13 +260,17 @@ media_channel_closed_cb (TpsipMediaChannel *chan, gpointer user_data)
 static TpsipMediaChannel *
 new_media_channel (TpsipMediaFactory *fac,
                    TpHandle initiator,
-                   TpHandle maybe_peer)
+                   TpHandle maybe_peer,
+                   TpsipMediaChannelCreationFlags flags)
 {
   TpsipMediaFactoryPrivate *priv;
   TpsipMediaChannel *chan = NULL;
   TpBaseConnection *conn;
   gchar *object_path;
   const gchar *nat_traversal = "none";
+  gboolean initial_audio;
+  gboolean initial_video;
+  gboolean immutable_streams = FALSE;
 
   g_assert (initiator != 0);
 
@@ -273,6 +282,13 @@ new_media_channel (TpsipMediaFactory *fac,
 
   DEBUG("channel object path %s", object_path);
 
+  initial_audio = ((flags & TPSIP_MEDIA_CHANNEL_CREATE_WITH_AUDIO) != 0);
+  initial_video = ((flags & TPSIP_MEDIA_CHANNEL_CREATE_WITH_VIDEO) != 0);
+
+  g_object_get (conn,
+      "immutable-streams", &immutable_streams,
+      NULL);
+
   if (priv->stun_server != NULL)
     {
       nat_traversal = "stun";
@@ -283,6 +299,9 @@ new_media_channel (TpsipMediaFactory *fac,
                        "object-path", object_path,
                        "handle", maybe_peer,
                        "initiator", initiator,
+                       "initial-audio", initial_audio,
+                       "initial-video", initial_video,
+                       "immutable-streams", immutable_streams,
                        "nat-traversal", nat_traversal,
                        NULL);
 
@@ -302,6 +321,16 @@ new_media_channel (TpsipMediaFactory *fac,
   return chan;
 }
 
+static void
+incoming_call_cb (TpsipMediaChannel *channel,
+                  TpsipMediaFactory *fac)
+{
+  g_signal_handlers_disconnect_by_func (channel,
+      G_CALLBACK (incoming_call_cb), fac);
+  tp_channel_manager_emit_new_channel (fac,
+      TP_EXPORTABLE_CHANNEL (channel), NULL);
+}
+
 static gboolean
 tpsip_nua_i_invite_cb (TpBaseConnection    *conn,
                        const TpsipNuaEvent *ev,
@@ -311,6 +340,7 @@ tpsip_nua_i_invite_cb (TpBaseConnection    *conn,
   TpsipMediaChannel *channel;
   TpHandleRepoIface *contact_repo;
   TpHandle handle;
+  guint channel_flags = 0;
 
   /* figure out a handle for the identity */
 
@@ -319,7 +349,7 @@ tpsip_nua_i_invite_cb (TpBaseConnection    *conn,
   handle = tpsip_handle_parse_from (contact_repo, ev->sip);
   if (!handle)
     {
-      g_message ("incoming INVITE with invalid sender information");
+      MESSAGE ("incoming INVITE with invalid sender information");
       nua_respond (ev->nua_handle, 400, "Invalid From address", TAG_END());
       return TRUE;
     }
@@ -327,14 +357,16 @@ tpsip_nua_i_invite_cb (TpBaseConnection    *conn,
   DEBUG("Got incoming invite from <%s>",
         tp_handle_inspect (contact_repo, handle));
 
-  channel = new_media_channel (fac, handle, handle);
+  channel = new_media_channel (fac, handle, handle, channel_flags);
 
   tpsip_media_channel_receive_invite (channel, ev->nua_handle);
 
   tp_handle_unref (contact_repo, handle);
 
-  tp_channel_manager_emit_new_channel (fac,
-      TP_EXPORTABLE_CHANNEL (channel), NULL);
+  /* We delay emission of NewChannel(s) until we have the data on
+   * initial media */
+  g_signal_connect (channel, "incoming-call",
+      G_CALLBACK (incoming_call_cb), fac);
 
   return TRUE;
 }
@@ -414,6 +446,8 @@ static const gchar * const media_channel_fixed_properties[] = {
 static const gchar * const named_channel_allowed_properties[] = {
     TP_IFACE_CHANNEL ".TargetHandle",
     TP_IFACE_CHANNEL ".TargetID",
+    TP_IFACE_CHANNEL_TYPE_STREAMED_MEDIA ".InitialAudio",
+    TP_IFACE_CHANNEL_TYPE_STREAMED_MEDIA ".InitialVideo",
     NULL
 };
 
@@ -468,6 +502,7 @@ tpsip_media_factory_requestotron (TpChannelManager *manager,
   TpsipMediaChannel *channel = NULL;
   GError *error = NULL;
   GSList *request_tokens;
+  guint chan_flags = 0;
   gboolean require_target_handle;
   gboolean add_peer_to_remote_pending;
 
@@ -532,7 +567,7 @@ tpsip_media_factory_requestotron (TpChannelManager *manager,
               &error))
         goto error;
 
-      channel = new_media_channel (self, conn->self_handle, 0);
+      channel = new_media_channel (self, conn->self_handle, 0, 0);
       break;
 
     case TP_HANDLE_TYPE_CONTACT:
@@ -543,6 +578,17 @@ tpsip_media_factory_requestotron (TpChannelManager *manager,
               &error))
         goto error;
 
+      /* Calls to self are problematic in terms of StreamedMedia channel
+       * interface and its semantically required Group member changes;
+       * we disable them until a better API is available through
+       * Call channel type */
+      if (handle == conn->self_handle)
+        {
+          g_set_error (&error, TP_ERRORS, TP_ERROR_NOT_IMPLEMENTED,
+              "Cannot call self");
+          goto error;
+        }
+
       if (method == METHOD_ENSURE)
         {
           guint i;
@@ -562,13 +608,23 @@ tpsip_media_factory_requestotron (TpChannelManager *manager,
             }
         }
 
-      channel = new_media_channel (self, conn->self_handle, handle);
+      if (tp_asv_get_boolean (request_properties,
+            TP_IFACE_CHANNEL_TYPE_STREAMED_MEDIA ".InitialAudio", NULL))
+        chan_flags |= TPSIP_MEDIA_CHANNEL_CREATE_WITH_AUDIO;
+
+      if (tp_asv_get_boolean (request_properties,
+            TP_IFACE_CHANNEL_TYPE_STREAMED_MEDIA ".InitialVideo", NULL))
+        chan_flags |= TPSIP_MEDIA_CHANNEL_CREATE_WITH_VIDEO;
+
+      channel = new_media_channel (self, conn->self_handle, handle, chan_flags);
 
       if (add_peer_to_remote_pending)
         {
           if (!_tpsip_media_channel_add_member ((GObject *) channel, handle,
                 "", &error))
             {
+              /* FIXME: do we really want to emit Closed in this case?
+               * There wasn't a NewChannel/NewChannels emission */
               tpsip_media_channel_close (channel);
               goto error;
             }
@@ -587,6 +643,8 @@ tpsip_media_factory_requestotron (TpChannelManager *manager,
       TP_EXPORTABLE_CHANNEL (channel), request_tokens);
   g_slist_free (request_tokens);
 
+  tpsip_media_channel_create_initial_streams (channel);
+
   return TRUE;
 
 error:
