@@ -1,7 +1,7 @@
 /*
  * sip-connection-helpers.c - Helper routines used by TpsipConnection
  * Copyright (C) 2005 Collabora Ltd.
- * Copyright (C) 2006-2008 Nokia Corporation
+ * Copyright (C) 2006-2009 Nokia Corporation
  *
  * This work is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -18,6 +18,8 @@
  * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
+#include <config.h>
+
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
@@ -28,6 +30,8 @@
 #include <telepathy-glib/interfaces.h>
 #include <telepathy-glib/svc-connection.h>
 
+#include <tpsip/util.h>
+
 #include "sip-connection-helpers.h"
 
 #include <sofia-sip/sip.h>
@@ -78,7 +82,20 @@ priv_sip_from_url_make (TpsipConnection *conn,
 
   g_object_get (conn, "alias", &alias, NULL);
   if (alias != NULL)
-    from->a_display = su_strdup (home, alias);
+    {
+      gchar *alias_quoted;
+
+      /* Make the alias into a quoted string, escaping all characters
+       * that cannot go verbatim into a quoted string */
+
+      alias_quoted = tpsip_quote_string (alias);
+
+      g_free (alias);
+
+      from->a_display = su_strdup (home, alias_quoted);
+
+      g_free (alias_quoted);
+    }
 
   return from;
 }
@@ -170,7 +187,7 @@ tpsip_conn_update_proxy_and_transport (TpsipConnection *conn)
           else if (g_ascii_strcasecmp (priv->transport, "udp") == 0)
             params = "transport=udp";
           else
-            g_warning ("unrecognized transport parameter value: %s", priv->transport);
+            WARNING ("unrecognized transport parameter value: %s", priv->transport);
         }
 
       nua_set_params (priv->sofia_nua,
@@ -206,7 +223,7 @@ tpsip_conn_get_local_url (TpsipConnection *conn)
     }
 
   if (priv->local_ip_address == NULL)
-    url->url_host = "*";
+    url->url_host = "0";
   else
     url->url_host = priv->local_ip_address;
 
@@ -249,7 +266,7 @@ priv_nua_get_outbound_options (nua_t* nua)
   if (outbound == NULL)
     return option_table;
 
-  g_debug ("%s: got outbound options %s", G_STRFUNC, outbound);
+  DEBUG ("got outbound options %s", outbound);
 
   options = g_strsplit_set (outbound, " ", 0);
 
@@ -321,7 +338,7 @@ priv_nua_set_outbound_options (nua_t* nua, GHashTable* option_table)
   g_assert (outbound != NULL);
 
   /* deliver the option string to the stack */
-  g_debug ("%s: setting outbound options %s", G_STRFUNC, outbound);
+  DEBUG ("setting outbound options %s", outbound);
   nua_set_params (nua, NUTAG_OUTBOUND(outbound), TAG_NULL());
 
   g_free (outbound);
@@ -389,7 +406,7 @@ priv_sanitize_keepalive_interval (TpsipConnectionPrivate *priv)
               : TPSIP_CONNECTION_MINIMUM_KEEPALIVE_INTERVAL;
       if (priv->keepalive_interval < minimum_interval)
         {
-          g_warning ("keepalive interval is too low, pushing to %u", minimum_interval);
+          WARNING ("keepalive interval is too low, pushing to %u", minimum_interval);
           priv->keepalive_interval = minimum_interval;
         }
     }
@@ -489,7 +506,7 @@ priv_stun_resolver_cb (sres_context_t *ctx, sres_query_t *query, sres_record_t *
     tpsip_conn_set_stun_server_address (conn,
                                         inet_ntoa (ans->a_addr));
   else
-    g_debug ("Couldn't resolv STUN server address, ignoring.");
+    DEBUG ("Couldn't resolv STUN server address, ignoring.");
 
   sres_free_answers (priv->sofia_resolver, answers);
 }
@@ -781,7 +798,7 @@ tpsip_handle_normalize (TpHandleRepoIface *repo,
 
       if (base_url == NULL || base_url->url_host == NULL)
         {
-          g_warning ("bare name given, but no account URL is set");
+          WARNING ("bare name given, but no account URL is set");
           goto error;
         }
 
@@ -933,3 +950,109 @@ tpsip_handle_parse_from (TpHandleRepoIface *contact_repo,
 
   return handle;
 }
+
+#ifdef HAVE_LIBIPHB
+
+static int
+heartbeat_wakeup (su_root_magic_t *foo,
+                  su_wait_t *wait,
+                  void *user_data)
+{
+  TpsipConnection *self = (TpsipConnection *) user_data;
+  TpsipConnectionPrivate *priv = TPSIP_CONNECTION_GET_PRIVATE (self);
+  gint keepalive_earliest; 
+
+  DEBUG("tick");
+
+  g_assert (priv->heartbeat != NULL);
+
+  if ((wait->revents & (SU_WAIT_IN | SU_WAIT_HUP | SU_WAIT_ERR)) != SU_WAIT_IN)
+    {
+      WARNING ("heartbeat descriptor invalidated prematurely with event mask %hd", wait->revents);
+      tpsip_conn_heartbeat_shutdown (self);
+      return 0;
+    }
+
+  keepalive_earliest = (int) priv->keepalive_interval - TPSIP_DEFER_TIMEOUT;
+  if (keepalive_earliest < 0)
+    keepalive_earliest = 0;
+
+  if (iphb_wait (priv->heartbeat,
+        (gushort) keepalive_earliest,
+        (gushort) MIN(priv->keepalive_interval, G_MAXUSHORT),
+        0) < 0)
+    {
+      WARNING ("iphb_wait failed");
+      tpsip_conn_heartbeat_shutdown (self);
+      return 0;
+    }
+
+  return 0;
+}
+
+#endif /* HAVE_LIBIPHB */
+
+void
+tpsip_conn_heartbeat_init (TpsipConnection *self)
+{
+#ifdef HAVE_LIBIPHB
+  TpsipConnectionPrivate *priv = TPSIP_CONNECTION_GET_PRIVATE (self);
+  int wait_id;
+  int reference_interval = 0;
+
+  g_assert (priv->heartbeat == NULL);
+
+  priv->heartbeat = iphb_open (&reference_interval);
+
+  if (priv->heartbeat == NULL)
+    {
+      WARNING ("opening IP heartbeat failed: %s", strerror (errno));
+      return;
+    }
+
+  DEBUG("heartbeat opened with reference interval %d", reference_interval);
+
+  su_wait_init (priv->heartbeat_wait);
+  if (su_wait_create (priv->heartbeat_wait,
+                      iphb_get_fd (priv->heartbeat),
+                      SU_WAIT_IN) != 0)
+    tpsip_log (DEBUG_FLAG, G_LOG_LEVEL_CRITICAL,
+        "could not create a wait object");
+
+  wait_id = su_root_register (priv->sofia_root,
+      priv->heartbeat_wait, heartbeat_wakeup, self, 0);
+
+  g_return_if_fail (wait_id > 0);
+  priv->heartbeat_wait_id = wait_id;
+
+  /* Prime the heartbeat for the first time.
+   * The minimum wakeup timeout is 0 to fall in step with other
+   * clients using the same interval */
+  if (iphb_wait (priv->heartbeat,
+        0, (gushort) MIN(priv->keepalive_interval, G_MAXUSHORT), 0) < 0)
+    {
+      WARNING ("iphb_wait failed");
+      tpsip_conn_heartbeat_shutdown (self);
+    }
+
+#endif /* HAVE_LIBIPHB */
+}
+
+void
+tpsip_conn_heartbeat_shutdown (TpsipConnection *self)
+{
+#ifdef HAVE_LIBIPHB
+  TpsipConnectionPrivate *priv = TPSIP_CONNECTION_GET_PRIVATE (self);
+
+  if (priv->heartbeat_wait_id == 0)
+    return;
+
+  su_root_deregister (priv->sofia_root, priv->heartbeat_wait_id);
+  priv->heartbeat_wait_id = 0;
+
+  su_wait_destroy (priv->heartbeat_wait);
+
+  iphb_close (priv->heartbeat);
+  priv->heartbeat = NULL;
+#endif /* HAVE_LIBIPHB */
+}
