@@ -1,7 +1,7 @@
 /*
  * sip-connection-helpers.h - Helper routines used by TpsipConnection
  * Copyright (C) 2005 Collabora Ltd.
- * Copyright (C) 2005-2008 Nokia Corporation
+ * Copyright (C) 2005-2009 Nokia Corporation
  *
  * This work is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -72,6 +72,13 @@ const url_t* tpsip_conn_get_contact_url (TpsipConnection *conn,
 TpHandle tpsip_handle_parse_from (TpHandleRepoIface *contact_repo,
                                   const sip_t *sip);
 
+/***********************************************************************
+ * Heartbeat management for keepalives
+ ***********************************************************************/
+
+void tpsip_conn_heartbeat_init (TpsipConnection *self);
+void tpsip_conn_heartbeat_shutdown (TpsipConnection *self);
+
 G_END_DECLS
 
 #endif /* #ifndef __TPSIP_CONNECTION_HELPERS_H__*/
