@@ -1,7 +1,7 @@
 /*
  * sip-connection.c - Source for TpsipConnection
  * Copyright (C) 2005-2007 Collabora Ltd.
- * Copyright (C) 2005-2008 Nokia Corporation
+ * Copyright (C) 2005-2009 Nokia Corporation
  *   @author Kai Vehmanen <first.surname@nokia.com>
  *   @author Martti Mela <first.surname@nokia.com>
  *   @author Mikhail Zabaluev <mikhail.zabaluev@nokia.com>
@@ -97,6 +97,7 @@ enum
   PROP_STUN_SERVER,        /**< STUN server address (if not set, derived
 			        from public SIP address */
   PROP_STUN_PORT,          /**< STUN port */
+  PROP_IMMUTABLE_STREAMS,  /**< If the session content is immutable once set up */
   PROP_LOCAL_IP_ADDRESS,   /**< Local IP address (normally not needed, chosen by stack) */
   PROP_LOCAL_PORT,         /**< Local port for SIP (normally not needed, chosen by stack) */
   PROP_EXTRA_AUTH_USER,	   /**< User name to use for extra authentication challenges */
@@ -234,7 +235,7 @@ tpsip_connection_set_property (GObject      *object,
       {
         if (mech != TPSIP_CONNECTION_KEEPALIVE_NONE
             && mech != TPSIP_CONNECTION_KEEPALIVE_AUTO)
-          g_warning ("keep-alive mechanism selection is ignored when interval is 0");
+          WARNING ("keep-alive mechanism selection is ignored when interval is 0");
       }
     else
       {
@@ -280,6 +281,9 @@ tpsip_connection_set_property (GObject      *object,
     priv->stun_host = g_value_dup_string (value);
     break;
   }
+  case PROP_IMMUTABLE_STREAMS:
+    priv->immutable_streams = g_value_get_boolean (value);
+    break;
   case PROP_LOCAL_IP_ADDRESS: {
     g_free (priv->local_ip_address);
     priv->local_ip_address = g_value_dup_string (value);
@@ -377,6 +381,9 @@ tpsip_connection_get_property (GObject      *object,
     g_value_set_uint (value, priv->stun_port);
     break;
   }
+  case PROP_IMMUTABLE_STREAMS:
+    g_value_set_boolean (value, priv->immutable_streams);
+    break;
   case PROP_LOCAL_IP_ADDRESS: {
     g_value_set_string (value, priv->local_ip_address);
     break;
@@ -545,6 +552,13 @@ tpsip_connection_class_init (TpsipConnectionClass *klass)
       G_PARAM_CONSTRUCT | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   INST_PROP(PROP_STUN_PORT);
 
+  param_spec = g_param_spec_boolean ("immutable-streams", "Immutable streams",
+      "Set if additional streams cannot be requested on a media channel,"
+      " nor existing streams can be removed",
+      FALSE,
+      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
+  INST_PROP(PROP_IMMUTABLE_STREAMS);
+
   param_spec = g_param_spec_string ("local-ip-address", "Local IP address",
       "Local IP address to use",
       NULL,
@@ -617,7 +631,7 @@ priv_handle_auth (TpsipConnection* self,
 
   if (realm == NULL)
     {
-      g_warning ("no realm presented for authentication");
+      WARNING ("no realm presented for authentication");
       return FALSE;
     }
 
@@ -629,7 +643,7 @@ priv_handle_auth (TpsipConnection* self,
         priv->registrar_realm = g_strdup (realm);
       else if (wa && strcmp(priv->registrar_realm, realm) != 0)
         {
-          g_message ("registrar realm changed from %s to %s", priv->registrar_realm, realm);
+          MESSAGE ("registrar realm changed from %s to %s", priv->registrar_realm, realm);
           g_free (priv->registrar_realm);
           priv->registrar_realm = g_strdup (realm);
         }
@@ -685,7 +699,7 @@ priv_handle_auth (TpsipConnection* self,
 
   if (auth == NULL)
     {
-      g_warning ("authentication data are incomplete");
+      WARNING ("authentication data are incomplete");
       return FALSE;
     }
 
@@ -727,6 +741,7 @@ tpsip_connection_nua_r_register_cb (TpsipConnection     *self,
                                     tagi_t               tags[],
                                     gpointer             foo)
 {
+  TpBaseConnection *base = (TpBaseConnection *) self;
   TpConnectionStatus conn_status = TP_CONNECTION_STATUS_DISCONNECTED;
   TpConnectionStatusReason reason = 0;
 
@@ -755,14 +770,18 @@ tpsip_connection_nua_r_register_cb (TpsipConnection     *self,
         }
       else /* if (ev->status == 200) */
         {
+          if (base->status != TP_CONNECTION_STATUS_CONNECTING)
+            return TRUE;
+
           DEBUG("succesfully registered to the network");
           conn_status = TP_CONNECTION_STATUS_CONNECTED;
           reason = TP_CONNECTION_STATUS_REASON_REQUESTED;
+
+          tpsip_conn_heartbeat_init (self);
         }
     }
 
-  tp_base_connection_change_status ((TpBaseConnection *) self,
-                                    conn_status, reason);
+  tp_base_connection_change_status (base, conn_status, reason);
 
   return TRUE;
 }
@@ -780,6 +799,8 @@ tpsip_connection_shut_down (TpBaseConnection *base)
   /* We disposed of the REGISTER handle in the disconnected method */
   g_assert (priv->register_op == NULL);
 
+  tpsip_conn_heartbeat_shutdown (self);
+
   if (priv->sofia_nua != NULL)
     nua_shutdown (priv->sofia_nua);
 
