@@ -1,7 +1,7 @@
 /*
  * sip-media-stream.c - Source for TpsipMediaStream
  * Copyright (C) 2006 Collabora Ltd.
- * Copyright (C) 2006-2008 Nokia Corporation
+ * Copyright (C) 2006-2010 Nokia Corporation
  *   @author Kai Vehmanen <first.surname@nokia.com>
  *   @author Mikhail Zabaluev <mikhail.zabaluev@nokia.com>
  *
@@ -31,10 +31,15 @@
 #include <telepathy-glib/enums.h>
 #include <telepathy-glib/errors.h>
 #include <telepathy-glib/gtypes.h>
+#include <telepathy-glib/interfaces.h>
+#include <telepathy-glib/svc-generic.h>
 #include <telepathy-glib/svc-media-interfaces.h>
+#include <telepathy-glib/util.h>
 
 #include "config.h"
 
+#include <tpsip/codec-param-formats.h>
+
 #include "sip-media-stream.h"
 #include "sip-media-session.h"
 
@@ -55,8 +60,10 @@ G_DEFINE_TYPE_WITH_CODE(TpsipMediaStream,
     tpsip_media_stream,
     G_TYPE_OBJECT,
     G_IMPLEMENT_INTERFACE (TP_TYPE_SVC_MEDIA_STREAM_HANDLER,
-      stream_handler_iface_init)
-    )
+      stream_handler_iface_init);
+    G_IMPLEMENT_INTERFACE (TP_TYPE_SVC_DBUS_PROPERTIES,
+      tp_dbus_properties_mixin_iface_init);
+  )
 
 /* signal enum */
 enum
@@ -73,8 +80,6 @@ enum
 
 static guint signals[SIG_LAST_SIGNAL] = {0};
 
-static GRegex *fmtp_attr_regex = NULL;
-
 /* properties */
 enum
 {
@@ -86,9 +91,15 @@ enum
   PROP_DIRECTION,
   PROP_PENDING_SEND_FLAGS,
   PROP_HOLD_STATE,
+  PROP_CREATED_LOCALLY,
+  PROP_NAT_TRAVERSAL,
+  PROP_STUN_SERVERS,
+  PROP_RELAY_INFO,
   LAST_PROPERTY
 };
 
+static GPtrArray *tpsip_media_stream_relay_info_empty = NULL;
+
 /* private structure */
 typedef struct _TpsipMediaStreamPrivate TpsipMediaStreamPrivate;
 
@@ -102,6 +113,7 @@ struct _TpsipMediaStreamPrivate
   guint direction;                /* see gobj. prop. 'direction' */
   guint pending_send_flags;       /* see gobj. prop. 'pending-send-flags' */
   gboolean hold_state;            /* see gobj. prop. 'hold-state' */
+  gboolean created_locally;       /* see gobj. prop. 'created-locally' */
 
   gchar *stream_sdp;              /* SDP description of the stream */
 
@@ -119,11 +131,13 @@ struct _TpsipMediaStreamPrivate
   gboolean ready_received;              /* our ready method has been called */
   gboolean playing;                     /* stream set to playing */
   gboolean sending;                     /* stream set to sending */
+  gboolean pending_remote_receive;      /* TRUE if remote is to agree to receive media */
   gboolean native_cands_prepared;       /* all candidates discovered */
   gboolean native_codecs_prepared;      /* all codecs discovered */
   gboolean push_remote_cands_pending;   /* SetRemoteCandidates emission is pending */
   gboolean push_remote_codecs_pending;  /* SetRemoteCodecs emission is pending */
   gboolean codec_intersect_pending;     /* codec intersection is pending */
+  gboolean requested_hold_state;        /* hold state last requested from the stream handler */
   gboolean dispose_has_run;
 };
 
@@ -133,8 +147,7 @@ static void push_remote_codecs (TpsipMediaStream *stream);
 static void push_remote_candidates (TpsipMediaStream *stream);
 static void push_active_candidate_pair (TpsipMediaStream *stream);
 static void priv_update_sending (TpsipMediaStream *stream,
-                                 TpMediaStreamDirection direction,
-                                 guint pending_send_flags);
+                                 TpMediaStreamDirection direction);
 static void priv_update_local_sdp(TpsipMediaStream *stream);
 static void priv_generate_sdp (TpsipMediaStream *stream);
 
@@ -180,31 +193,42 @@ tpsip_media_stream_init (TpsipMediaStream *self)
   priv->push_remote_codecs_pending = FALSE;
 }
 
-static GObject *
-tpsip_media_stream_constructor (GType type, guint n_props,
-			      GObjectConstructParam *props)
+static void
+tpsip_media_stream_constructed (GObject *obj)
 {
-  GObject *obj;
-  TpsipMediaStreamPrivate *priv;
+  TpsipMediaStreamPrivate *priv = TPSIP_MEDIA_STREAM_GET_PRIVATE (
+      TPSIP_MEDIA_STREAM (obj));
+  GObjectClass *parent_object_class =
+      G_OBJECT_CLASS (tpsip_media_stream_parent_class);
   DBusGConnection *bus;
 
-  /* call base class constructor */
-  obj = G_OBJECT_CLASS (tpsip_media_stream_parent_class)->
-           constructor (type, n_props, props);
-  priv = TPSIP_MEDIA_STREAM_GET_PRIVATE (TPSIP_MEDIA_STREAM (obj));
+  /* call base class method */
+  if (parent_object_class->constructed != NULL)
+    parent_object_class->constructed (obj);
+
+  /* XXX: overloading the remote pending send flag to check
+   * if this is a locally offered stream. The code creating such streams
+   * always sets the flag, because the remote end is supposed to decide
+   * whether it wants to send.
+   * This may look weird during a local hold. However, the pending flag
+   * will be harmlessly cleared once the offer-answer is complete. */
+  if ((priv->direction & TP_MEDIA_STREAM_DIRECTION_SEND) != 0
+      && (priv->pending_send_flags & TP_MEDIA_STREAM_PENDING_REMOTE_SEND) != 0)
+    {
+      /* Block sending until the stream is remotely accepted */
+      priv->pending_remote_receive = TRUE;
+    }
 
   /* go for the bus */
   bus = tp_get_bus ();
   dbus_g_connection_register_g_object (bus, priv->object_path, obj);
-
-  return obj;
 }
 
 static void
 tpsip_media_stream_get_property (GObject    *object,
-			       guint       property_id,
-			       GValue     *value,
-			       GParamSpec *pspec)
+			         guint       property_id,
+			         GValue     *value,
+			         GParamSpec *pspec)
 {
   TpsipMediaStream *stream = TPSIP_MEDIA_STREAM (object);
   TpsipMediaStreamPrivate *priv = TPSIP_MEDIA_STREAM_GET_PRIVATE (stream);
@@ -235,6 +259,19 @@ tpsip_media_stream_get_property (GObject    *object,
     case PROP_HOLD_STATE:
       g_value_set_boolean (value, priv->hold_state);
       break;
+    case PROP_CREATED_LOCALLY:
+      g_value_set_boolean (value, priv->created_locally);
+      break;
+    case PROP_NAT_TRAVERSAL:
+      g_value_set_static_string (value, "none");
+      break;
+    case PROP_STUN_SERVERS:
+      g_return_if_fail (priv->session != NULL);
+      g_object_get_property (G_OBJECT (priv->session), "stun-servers", value);
+      break;
+    case PROP_RELAY_INFO:
+      g_value_set_static_boxed (value, tpsip_media_stream_relay_info_empty);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
     }
@@ -242,9 +279,9 @@ tpsip_media_stream_get_property (GObject    *object,
 
 static void
 tpsip_media_stream_set_property (GObject      *object,
-			       guint         property_id,
-			       const GValue *value,
-			       GParamSpec   *pspec)
+			         guint         property_id,
+			         const GValue *value,
+			         GParamSpec   *pspec)
 {
   TpsipMediaStream *stream = TPSIP_MEDIA_STREAM (object);
   TpsipMediaStreamPrivate *priv = TPSIP_MEDIA_STREAM_GET_PRIVATE (stream);
@@ -276,6 +313,9 @@ tpsip_media_stream_set_property (GObject      *object,
     case PROP_HOLD_STATE:
       priv->hold_state = g_value_get_boolean (value);
       break;
+    case PROP_CREATED_LOCALLY:
+      priv->created_locally = g_value_get_boolean (value);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
     }
@@ -287,13 +327,30 @@ static void tpsip_media_stream_finalize (GObject *object);
 static void
 tpsip_media_stream_class_init (TpsipMediaStreamClass *klass)
 {
+  static TpDBusPropertiesMixinPropImpl stream_handler_props[] = {
+      { "CreatedLocally", "created-locally", NULL },
+      { "NATTraversal", "nat-traversal", NULL },
+      { "STUNServers", "stun-servers", NULL },
+      { "RelayInfo", "relay-info", NULL },
+      { NULL }
+  };
+
+  static TpDBusPropertiesMixinIfaceImpl prop_interfaces[] = {
+      { TP_IFACE_MEDIA_STREAM_HANDLER,
+        tp_dbus_properties_mixin_getter_gobject_properties,
+        NULL,
+        stream_handler_props,
+      },
+      { NULL }
+  };
+
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
   GType stream_type = G_OBJECT_CLASS_TYPE (klass);
   GParamSpec *param_spec;
 
   g_type_class_add_private (klass, sizeof (TpsipMediaStreamPrivate));
 
-  object_class->constructor = tpsip_media_stream_constructor;
+  object_class->constructed = tpsip_media_stream_constructed;
 
   object_class->get_property = tpsip_media_stream_get_property;
   object_class->set_property = tpsip_media_stream_set_property;
@@ -301,7 +358,7 @@ tpsip_media_stream_class_init (TpsipMediaStreamClass *klass)
   object_class->dispose = tpsip_media_stream_dispose;
   object_class->finalize = tpsip_media_stream_finalize;
 
-  param_spec = g_param_spec_object ("media-session", "GabbleMediaSession object",
+  param_spec = g_param_spec_object ("media-session", "TpsipMediaSession object",
       "SIP media session object that owns this media stream object.",
       TPSIP_TYPE_MEDIA_SESSION,
       G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
@@ -362,6 +419,32 @@ tpsip_media_stream_class_init (TpsipMediaStreamClass *klass)
                                    PROP_HOLD_STATE,
                                    param_spec);
 
+  param_spec = g_param_spec_boolean ("created-locally", "Created locally?",
+      "True if this stream was created by the local user", FALSE,
+      G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class, PROP_CREATED_LOCALLY,
+      param_spec);
+
+  param_spec = g_param_spec_string ("nat-traversal", "NAT traversal",
+      "NAT traversal mechanism for this stream", NULL,
+      G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class, PROP_NAT_TRAVERSAL,
+      param_spec);
+
+  param_spec = g_param_spec_boxed ("stun-servers", "STUN servers",
+      "Array of IP address-port pairs for available STUN servers",
+      TP_ARRAY_TYPE_SOCKET_ADDRESS_IP_LIST,
+      G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class, PROP_STUN_SERVERS, param_spec);
+
+  param_spec = g_param_spec_boxed ("relay-info", "Relay info",
+      "Array of mappings containing relay server information",
+      TP_ARRAY_TYPE_STRING_VARIANT_MAP_LIST,
+      G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class, PROP_RELAY_INFO, param_spec);
+
+  tpsip_media_stream_relay_info_empty = g_ptr_array_new ();
+
   /* signals not exported by DBus interface */
   signals[SIG_READY] =
     g_signal_new ("ready",
@@ -417,20 +500,9 @@ tpsip_media_stream_class_init (TpsipMediaStreamClass *klass)
                   g_cclosure_marshal_VOID__VOID,
                   G_TYPE_NONE, 0);
 
-/* Regexps for the name and the value parts of the parameter syntax */
-#define FMTP_TOKEN_ATTR "([-A-Za-z0-9_.]+)"
-#define FMTP_TOKEN_VALUE "([^;\"\\s]+|\"([^\"\\\\]|\\\\.)*\")"
-/* Indexes of the respective match groups in the whole regexp below */
-#define FMTP_MATCH_ATTR 1
-#define FMTP_MATCH_VALUE 2
-
-  fmtp_attr_regex = g_regex_new (FMTP_TOKEN_ATTR
-                                   "\\s*=\\s*"
-                                   FMTP_TOKEN_VALUE,
-                                 G_REGEX_RAW,
-                                 0 /* G_REGEX_MATCH_ANCHORED */,
-                                 NULL);
-  g_assert (fmtp_attr_regex != NULL);
+  klass->dbus_props_class.interfaces = prop_interfaces;
+  tp_dbus_properties_mixin_class_init (object_class,
+      G_STRUCT_OFFSET (TpsipMediaStreamClass, dbus_props_class));
 }
 
 void
@@ -491,7 +563,7 @@ tpsip_media_stream_codec_choice (TpSvcMediaStreamHandler *iface,
   /* Inform the connection manager of the current codec choice. 
    * -> note: not implemented by tp-gabble either (2006/May) */
 
-  g_debug ("%s: not implemented (ignoring)", G_STRFUNC);
+  DEBUG ("not implemented (ignoring)");
 
   tp_svc_media_stream_handler_return_from_codec_choice (context);
 }
@@ -607,7 +679,7 @@ tpsip_media_stream_new_native_candidate (TpSvcMediaStreamHandler *iface,
 
   if (priv->stream_sdp != NULL)
     {
-      g_message ("Stream %u: SDP already generated, ignoring native candidate '%s'", priv->id, candidate_id);
+      MESSAGE ("Stream %u: SDP already generated, ignoring native candidate '%s'", priv->id, candidate_id);
       tp_svc_media_stream_handler_return_from_new_native_candidate (context);
       return;
     }
@@ -669,6 +741,45 @@ priv_set_local_codecs (TpsipMediaStream *self,
     priv_generate_sdp (self);
 }
 
+static void
+tpsip_media_stream_codecs_updated (TpSvcMediaStreamHandler *iface,
+                                   const GPtrArray *codecs,
+                                   DBusGMethodInvocation *context)
+{
+  TpsipMediaStream *self = TPSIP_MEDIA_STREAM (iface);
+  TpsipMediaStreamPrivate *priv = TPSIP_MEDIA_STREAM_GET_PRIVATE (self);
+
+  if (!priv->native_codecs_prepared)
+    {
+      GError e = { TP_ERRORS, TP_ERROR_NOT_AVAILABLE,
+          "CodecsUpdated may not be called before codecs have been provided "
+          "with SetLocalCodecs or Ready" };
+
+      SESSION_DEBUG(priv->session,
+          "CodecsUpdated called before SetLocalCodecs or Ready");
+
+      dbus_g_method_return_error (context, &e);
+    }
+  else
+    {
+      GValue val = { 0, };
+
+      SESSION_DEBUG(priv->session, "putting list of %d locally supported "
+          "codecs from CodecsUpdated into cache", codecs->len);
+      g_value_init (&val, TP_ARRAY_TYPE_MEDIA_STREAM_HANDLER_CODEC_LIST);
+      g_value_set_static_boxed (&val, codecs);
+      g_value_copy (&val, &priv->native_codecs);
+
+      /* This doesn't use priv_generate_sdp because it short-circuits if
+       * priv->stream_sdp is already set. We want to update it.
+       */
+      if (priv->native_cands_prepared)
+        priv_update_local_sdp (self);
+
+      tp_svc_media_stream_handler_return_from_codecs_updated (context);
+    }
+}
+
 /**
  * tpsip_media_stream_ready
  *
@@ -697,7 +808,7 @@ tpsip_media_stream_ready (TpSvcMediaStreamHandler *iface,
 
   if (priv->ready_received)
     {
-      g_message ("Ready called more than once");
+      MESSAGE ("Ready called more than once");
       tp_svc_media_stream_handler_return_from_ready (context);
       return;
     }
@@ -823,7 +934,7 @@ tpsip_media_stream_supported_codecs (TpSvcMediaStreamHandler *iface,
         }
     }
   else
-    g_warning("SupportedCodecs called when no intersection is ongoing");
+    WARNING("SupportedCodecs called when no intersection is ongoing");
 
   tp_svc_media_stream_handler_return_from_supported_codecs (context);
 }
@@ -841,6 +952,10 @@ static void
 tpsip_media_stream_unhold_failure (TpSvcMediaStreamHandler *self,
                                    DBusGMethodInvocation *context)
 {
+  /* Not doing anything to hold_state or requested_hold_state,
+   * because the session is going to put all streams on hold after getting
+   * the signal below */
+
   g_signal_emit (self, signals[SIG_UNHOLD_FAILURE], 0);
   tp_svc_media_stream_handler_return_from_unhold_failure (context);
 }
@@ -880,9 +995,10 @@ const char *tpsip_media_stream_local_sdp (TpsipMediaStream *obj)
   return priv->stream_sdp;
 }
 
-static inline guint
-tpsip_tp_stream_direction_from_remote (sdp_mode_t mode)
+TpMediaStreamDirection
+tpsip_media_stream_direction_from_remote_media (const sdp_media_t *media)
 {
+  sdp_mode_t mode = media->m_mode;
   return ((mode & sdp_recvonly)? TP_MEDIA_STREAM_DIRECTION_SEND : 0)
        | ((mode & sdp_sendonly)? TP_MEDIA_STREAM_DIRECTION_RECEIVE : 0);
 }
@@ -910,7 +1026,7 @@ priv_get_requested_direction (TpsipMediaStreamPrivate *priv)
   TpMediaStreamDirection direction;
 
   direction = priv->direction;
-  if (priv->pending_send_flags != 0)
+  if ((priv->pending_send_flags & TP_MEDIA_STREAM_PENDING_LOCAL_SEND) != 0)
     direction |= TP_MEDIA_STREAM_DIRECTION_SEND;
   return direction;
 }
@@ -960,20 +1076,20 @@ tpsip_media_stream_set_remote_media (TpsipMediaStream *stream,
 
   if (new_media->m_proto != sdp_proto_rtp)
     {
-      g_warning ("Stream %u: the remote protocol is not RTP/AVP", priv->id);
+      WARNING ("Stream %u: the remote protocol is not RTP/AVP", priv->id);
       return FALSE;
     }
 
   sdp_conn = sdp_media_connections (new_media);
   if (sdp_conn == NULL)
     {
-      g_warning ("Stream %u: no valid remote connections", priv->id);
+      WARNING ("Stream %u: no valid remote connections", priv->id);
       return FALSE;
     }
 
   if (new_media->m_rtpmaps == NULL)
     {
-      g_warning ("Stream %u: no remote codecs", priv->id);
+      WARNING ("Stream %u: no remote codecs", priv->id);
       return FALSE;
     }
 
@@ -991,7 +1107,7 @@ tpsip_media_stream_set_remote_media (TpsipMediaStream *stream,
     }
 
   old_direction = priv_get_requested_direction (priv);
-  new_direction = tpsip_tp_stream_direction_from_remote (new_media->m_mode);
+  new_direction = tpsip_media_stream_direction_from_remote_media (new_media);
 
   /* Make sure the peer can only enable sending or receiving direction
    * if it's allowed to */
@@ -1010,8 +1126,7 @@ tpsip_media_stream_set_remote_media (TpsipMediaStream *stream,
       /* Disable sending at this point if it will be disabled
        * accordingly to the new direction */
       priv_update_sending (stream,
-                           priv->direction & new_direction,
-                           priv->pending_send_flags);
+                           priv->direction & new_direction);
     }
 
   /* First add the new candidate, then update the codec set.
@@ -1118,28 +1233,21 @@ tpsip_media_stream_set_sending (TpsipMediaStream *stream, gboolean sending)
 
 static void
 priv_update_sending (TpsipMediaStream *stream,
-                     TpMediaStreamDirection direction,
-                     guint pending_send_flags)
+                     TpMediaStreamDirection direction)
 {
+  TpsipMediaStreamPrivate *priv = TPSIP_MEDIA_STREAM_GET_PRIVATE (stream);
   gboolean sending = TRUE;
 
   /* XXX: the pending send flag check is probably an overkill
    * considering that effective sending direction and pending send should be
    * mutually exclusive */
   if ((direction & TP_MEDIA_STREAM_DIRECTION_SEND) == 0
-      || (pending_send_flags
-          & (TP_MEDIA_STREAM_PENDING_REMOTE_SEND
-             | TP_MEDIA_STREAM_PENDING_LOCAL_SEND))
-         != 0)
+      || priv->pending_remote_receive
+      || (priv->pending_send_flags & TP_MEDIA_STREAM_PENDING_LOCAL_SEND) != 0
+      || !tpsip_media_session_is_accepted (priv->session))
     {
       sending = FALSE;
     }
-  else
-    {
-      TpsipMediaStreamPrivate *priv = TPSIP_MEDIA_STREAM_GET_PRIVATE (stream);
-      if (!tpsip_media_session_is_accepted (priv->session))
-        sending = FALSE;
-    }
 
   tpsip_media_stream_set_sending (stream, sending);
 }
@@ -1156,14 +1264,36 @@ tpsip_media_stream_set_direction (TpsipMediaStream *stream,
   priv = TPSIP_MEDIA_STREAM_GET_PRIVATE (stream);
   pending_send_flags = priv->pending_send_flags & pending_send_mask;
 
-  if ((direction & ~priv->direction & TP_MEDIA_STREAM_DIRECTION_SEND) != 0
-      && (priv->pending_send_flags & ~pending_send_mask) == 0)
+  if ((direction & ~priv->direction & TP_MEDIA_STREAM_DIRECTION_SEND) != 0)
     {
-      /* We're requested to start sending, but we need to confirm this
-       * with the client or the remote party.
-       * Clear the sending bit and set the pending send flag. */
-      direction &= ~(guint)TP_MEDIA_STREAM_DIRECTION_SEND;
-      pending_send_flags |= pending_send_mask;
+      /* We are requested to start sending, but... */
+      if ((pending_send_mask
+            & TP_MEDIA_STREAM_PENDING_LOCAL_SEND) != 0)
+        {
+          /* ... but we need to confirm this with the client.
+           * Clear the sending bit and set the pending send flag. */
+          direction &= ~(guint)TP_MEDIA_STREAM_DIRECTION_SEND;
+          pending_send_flags |= TP_MEDIA_STREAM_PENDING_LOCAL_SEND;
+        }
+      if ((pending_send_mask
+              & TP_MEDIA_STREAM_PENDING_REMOTE_SEND) != 0
+          && (priv->pending_send_flags
+              & TP_MEDIA_STREAM_PENDING_LOCAL_SEND) == 0)
+        {
+          g_assert ((priv_get_requested_direction (priv) & TP_MEDIA_STREAM_DIRECTION_SEND) == 0);
+
+          /* ... but the caller wants to agree with the remote
+           * end first. Block the stream handler from sending for now. */
+          priv->pending_remote_receive = TRUE;
+        }
+    }
+  if ((direction & ~priv->direction & TP_MEDIA_STREAM_DIRECTION_RECEIVE) != 0
+      && (pending_send_mask
+          & TP_MEDIA_STREAM_PENDING_REMOTE_SEND) != 0)
+    {
+      /* We're requested to start receiving, but the remote end did not
+       * confirm if it will send. Set the pending send flag. */
+      pending_send_flags |= TP_MEDIA_STREAM_PENDING_REMOTE_SEND;
     }
 
   if (priv->direction == direction
@@ -1181,7 +1311,7 @@ tpsip_media_stream_set_direction (TpsipMediaStream *stream,
                  priv->direction, priv->pending_send_flags);
 
   if (priv->remote_media != NULL)
-    priv_update_sending (stream, priv->direction, priv->pending_send_flags);
+    priv_update_sending (stream, priv->direction);
 
   if (priv->native_cands_prepared
       && priv->native_codecs_prepared
@@ -1191,24 +1321,32 @@ tpsip_media_stream_set_direction (TpsipMediaStream *stream,
 }
 
 /*
- * Checks for the pending send flag(s) present in @pending_send_mask,
- * if set, enable the sending bit in the stream direction and clear
- * the flag(s) that have been checked.
+ * Clears the pending send flag(s) present in @pending_send_mask.
+ * If #TP_MEDIA_STREAM_PENDING_LOCAL_SEND is thus cleared,
+ * enable the sending bit in the stream direction.
+ * If @pending_send_mask has #TP_MEDIA_STREAM_PENDING_REMOTE_SEND flag set,
+ * also start sending if agreed by the stream direction.
  */
 void
-tpsip_media_stream_apply_pending_send (TpsipMediaStream *stream,
-                                       guint pending_send_mask)
+tpsip_media_stream_apply_pending_direction (TpsipMediaStream *stream,
+                                            guint pending_send_mask)
 {
   TpsipMediaStreamPrivate *priv = TPSIP_MEDIA_STREAM_GET_PRIVATE (stream);
+  guint flags;
+
 
   /* Don't apply pending send for new streams that haven't been negotiated */
   if (priv->remote_media == NULL)
     return;
 
-  if ((priv->pending_send_flags & pending_send_mask) != 0)
+  /* Remember the flags that got changes and then clear the set */
+  flags = (priv->pending_send_flags & pending_send_mask);
+  priv->pending_send_flags &= ~pending_send_mask;
+
+  if (flags != 0)
     {
-      priv->direction |= TP_MEDIA_STREAM_DIRECTION_SEND;
-      priv->pending_send_flags &= ~pending_send_mask;
+      if ((flags & TP_MEDIA_STREAM_PENDING_LOCAL_SEND) != 0)
+        priv->direction |= TP_MEDIA_STREAM_DIRECTION_SEND;
 
       DEBUG("set direction %u, pending send flags %u", priv->direction, priv->pending_send_flags);
 
@@ -1216,8 +1354,14 @@ tpsip_media_stream_apply_pending_send (TpsipMediaStream *stream,
                      priv->direction, priv->pending_send_flags);
     }
 
+  if ((pending_send_mask & TP_MEDIA_STREAM_PENDING_REMOTE_SEND) != 0)
+    {
+      priv->pending_remote_receive = FALSE;
+      DEBUG("remote end ready to receive");
+    }
+
   /* Always check to enable sending because the session could become accepted */
-  priv_update_sending (stream, priv->direction, priv->pending_send_flags);
+  priv_update_sending (stream, priv->direction);
 }
 
 TpMediaStreamDirection
@@ -1259,6 +1403,20 @@ tpsip_media_stream_stop_telephony_event  (TpsipMediaStream *self)
         (TpSvcMediaStreamHandler *)self);
 }
 
+gboolean
+tpsip_media_stream_request_hold_state (TpsipMediaStream *self, gboolean hold)
+{
+  TpsipMediaStreamPrivate *priv = TPSIP_MEDIA_STREAM_GET_PRIVATE (self);
+
+  if ((!priv->requested_hold_state) != (!hold))
+    {
+      priv->requested_hold_state = hold;
+      tp_svc_media_stream_handler_emit_set_stream_held (self, hold);
+      return TRUE;
+    }
+  return FALSE;
+}
+
 static void
 priv_generate_sdp (TpsipMediaStream *self)
 {
@@ -1274,116 +1432,6 @@ priv_generate_sdp (TpsipMediaStream *self)
   g_signal_emit (self, signals[SIG_READY], 0);
 }
 
-static gchar *
-priv_fmtp_fetch_value (const GMatchInfo *match)
-{
-  const gchar *str;
-  gchar *value;
-  gchar *pv;
-  gint vstart = 0;
-  gint vend = 0;
-  gint i;
-
-  str = g_match_info_get_string (match);
-  g_match_info_fetch_pos (match, FMTP_MATCH_VALUE, &vstart, &vend);
-
-  if (vstart < vend - 1 && str[vstart] == '"' && str[vend - 1] == '"')
-    {
-      /* Unescape the quoted pair contents */
-
-      /* The buffer should be enough for the non-collapsed contents
-       * minus the quotes */
-      value = g_malloc ((vend - 1 - vstart) * sizeof (gchar));
-      pv = value;
-      for (i = vstart + 1; i < vend - 1; i++)
-        {
-          if (str[i] == '\\')
-            ++i;
-          *pv++ = str[i];
-        }
-      *pv = '\0';
-    }
-  else
-    {
-      /* Pass as is */
-      value = g_strndup (str + vstart, vend - vstart);
-    }
-
-  return value;
-}
-
-static int
-priv_fmtp_match_length (const GMatchInfo *match)
-{
-  gint attr_start = 0;
-  gint attr_end = 0;
-
-  g_match_info_fetch_pos (match, 0, &attr_start, &attr_end);
-
-  g_assert (attr_start == 0);
-  g_assert (attr_end > 0);
-
-  return attr_end;
-}
-
-static void
-priv_fmtp_parse (const char *fmtp, GHashTable *param_hash)
-{
-  GMatchInfo *match;
-  gchar *name;
-  gchar *value;
-  const char *p = fmtp;
-
-  if (fmtp == NULL)
-    return;
-
-  for (;;)
-    {
-      while (g_ascii_isspace (*p))
-        ++p;
-
-      if (*p == '\0')
-        break;
-
-      g_regex_match (fmtp_attr_regex, p, G_REGEX_MATCH_ANCHORED, &match);
-      if (!g_match_info_matches (match))
-        goto pass_as_is;
-
-      name = g_match_info_fetch (match, FMTP_MATCH_ATTR);
-
-      value = priv_fmtp_fetch_value (match);
-
-      p += priv_fmtp_match_length (match);
-
-      g_match_info_free (match);
-
-      DEBUG("parsed parameter %s=%s", name, value);
-
-      g_hash_table_insert (param_hash, name, value);
-
-      while (g_ascii_isspace (*p))
-        ++p;
-
-      switch (*p)
-        {
-        case '\0':
-          break;
-        case ';':
-          ++p;
-          break;
-        default:
-          goto pass_as_is;
-        }
-    }
-
-  return;
-
-pass_as_is:
-  DEBUG("failed to parse format-specific parameters as an attribute-value list: %s", fmtp);
-  g_hash_table_remove_all (param_hash);
-  g_hash_table_insert (param_hash, g_strdup (""), g_strdup (fmtp));
-}
-
 /**
  * Notify StreamEngine of remote codecs.
  *
@@ -1398,6 +1446,8 @@ static void push_remote_codecs (TpsipMediaStream *stream)
   GType codec_type;
   const sdp_media_t *sdpmedia;
   const sdp_rtpmap_t *rtpmap;
+  gchar *ptime = NULL;
+  gchar *max_ptime = NULL;
 
   DEBUG ("enter");
 
@@ -1417,6 +1467,21 @@ static void push_remote_codecs (TpsipMediaStream *stream)
       return;
     }
 
+  ptime = tpsip_sdp_get_string_attribute (sdpmedia->m_attributes, "ptime");
+  if (ptime == NULL)
+    {
+      g_object_get (priv->session,
+          "remote-ptime", &ptime,
+          NULL);
+    }
+  max_ptime = tpsip_sdp_get_string_attribute (sdpmedia->m_attributes, "maxptime");
+  if (max_ptime == NULL)
+    {
+      g_object_get (priv->session,
+          "remote-max-ptime", &max_ptime,
+          NULL);
+    }
+
   codec_type = TP_STRUCT_TYPE_MEDIA_STREAM_HANDLER_CODEC;
   codecs_type = TP_ARRAY_TYPE_MEDIA_STREAM_HANDLER_CODEC_LIST;
 
@@ -1435,7 +1500,15 @@ static void push_remote_codecs (TpsipMediaStream *stream)
       g_value_take_boxed (&codec,
                           dbus_g_type_specialized_construct (codec_type));
 
-      priv_fmtp_parse (rtpmap->rm_fmtp, opt_params);
+      if (ptime != NULL)
+        g_hash_table_insert (opt_params,
+            g_strdup("ptime"), g_strdup (ptime));
+      if (max_ptime != NULL)
+        g_hash_table_insert (opt_params,
+            g_strdup("maxptime"), g_strdup (max_ptime));
+
+      tpsip_codec_param_parse (priv->media_type, rtpmap->rm_encoding,
+          rtpmap->rm_fmtp, opt_params);
 
       /* RFC2327: see "m=" line definition 
        *  - note, 'encoding_params' is assumed to be channel
@@ -1464,6 +1537,8 @@ static void push_remote_codecs (TpsipMediaStream *stream)
     }
 
   g_hash_table_destroy (opt_params);
+  g_free (ptime);
+  g_free (max_ptime);
 
   SESSION_DEBUG(priv->session, "passing %d remote codecs to stream engine",
                 codecs->len);
@@ -1649,33 +1724,13 @@ switch (media_type)
 return "-";
 }
 
-static void
-priv_marshal_param (gpointer key,
-                    gpointer val,
-                    gpointer user_data)
-{
-  GString *params = (GString *) user_data;
-  const char *name = (const char *) key;
-  const char *value = (const char *) val;
-
-  if (params->len != 0)
-    g_string_append_c (params, ';');
-
-  if (name == NULL || !*name)
-    g_string_append_printf (params, "%s", value);
-  else if (value == NULL)
-    g_string_append_printf (params, "%s", name);
-  else
-    g_string_append_printf (params, "%s=%s", name, value);
-}
-
 static void
 priv_append_rtpmaps (const GPtrArray *codecs, GString *mline, GString *alines)
 {
   GValue codec = { 0, };
   gchar *co_name = NULL;
   guint co_id;
-  /* guint co_type; */
+  guint co_type;
   guint co_clockrate;
   guint co_channels;
   GHashTable *co_params = NULL;
@@ -1690,7 +1745,7 @@ priv_append_rtpmaps (const GPtrArray *codecs, GString *mline, GString *alines)
       dbus_g_type_struct_get (&codec,
                               0, &co_id,
                               1, &co_name,
-                              /* 2, &co_type, */
+                              2, &co_type,
                               3, &co_clockrate,
                               4, &co_channels,
                               5, &co_params,
@@ -1714,7 +1769,8 @@ priv_append_rtpmaps (const GPtrArray *codecs, GString *mline, GString *alines)
           GString *fmtp_value;
           g_string_append_printf (alines, "a=fmtp:%u ", co_id);
           fmtp_value = g_string_new (NULL);
-          g_hash_table_foreach (co_params, priv_marshal_param, fmtp_value);
+          tpsip_codec_param_format (co_type, co_name,
+              co_params, fmtp_value);
           g_string_append (alines, fmtp_value->str);
           g_string_free (fmtp_value, TRUE);
           g_string_append (alines, "\r\n");
@@ -1781,7 +1837,7 @@ priv_update_local_sdp(TpsipMediaStream *stream)
 
       if (ca_tports->len == 0)
         {
-          g_warning ("candidate '%s' lists no transports, skipping", candidate_id);
+          WARNING ("candidate '%s' lists no transports, skipping", candidate_id);
           continue;
         }
 
@@ -1922,6 +1978,7 @@ stream_handler_iface_init (gpointer g_iface, gpointer iface_data)
   IMPLEMENT(new_native_candidate);
   IMPLEMENT(ready);
   IMPLEMENT(set_local_codecs);
+  IMPLEMENT(codecs_updated);
   IMPLEMENT(stream_state);
   IMPLEMENT(supported_codecs);
   IMPLEMENT(hold_state);
