@@ -1,7 +1,7 @@
 /*
  * text-factory.c - Text channel factory for SIP connection manager
  * Copyright (C) 2007-2008 Collabora Ltd.
- * Copyright (C) 2007-2008 Nokia Corporation
+ * Copyright (C) 2007-2009 Nokia Corporation
  *
  * This work is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -286,7 +286,7 @@ tpsip_text_factory_new_channel (TpsipTextFactory *fac,
   object_path = g_strdup_printf ("%s/TextChannel%u",
       conn->object_path, handle);
 
-  g_debug ("%s: object path %s", G_STRFUNC, object_path);
+  DEBUG ("object path %s", object_path);
 
   chan = g_object_new (TPSIP_TYPE_TEXT_CHANNEL,
                        "connection", priv->conn,
@@ -506,7 +506,7 @@ tpsip_nua_i_message_cb (TpBaseConnection    *conn,
               gint status;
               const char *message = NULL;
 
-              g_message ("character set conversion failed for the message body: %s", error->message);
+              MESSAGE ("character set conversion failed for the message body: %s", error->message);
 
               if (error->code == G_CONVERT_ERROR_ILLEGAL_SEQUENCE)
                 {
